"""
This class inherits from `Experiment`, and also holds the name of the variable to profile
"""

import logging
import multiprocessing as mp
import os

import h5py
import numpy as np
from scipy.special import erfcinv

from freqfit.experiment import Experiment

NUM_CORES = 20  # TODO: change this to an environment variable, or something that detects available cores
SEED = 42

log = logging.getLogger(__name__)


class SetLimit(Experiment):
    def __init__(
        self,
        config: dict,
        jobid: int = 0,
        numtoy: int = 0,
        out_path: str = ".",
        name: str = "",
        overwrite_files: bool = False,
        numcores: int = NUM_CORES,
        profile_grid: np.array = None,
    ) -> None:
        """
        This class inherits from `Experiment`, and also holds the name of the variable to profile
        """
        super().__init__(config, name)

        self.test_statistics = (
            None  # maybe we want to store the test statistics internally?
        )
        self.var_to_profile = None
        self.jobid = jobid
        self.numtoy = numtoy
        self.out_path = out_path
        self.numcores = numcores
        self.overwrite_files = overwrite_files
        self.profile_grid = profile_grid

    def set_var_to_profile(self, var_to_profile: str):
        """
        Parameters
        ----------
        var_to_profile
            string -- the variable we are going to scan over to compute test-statistics at
        """

        if var_to_profile not in self.fitparameters:
            msg = f"variable '{var_to_profile}' not found in fit parameters"
            logging.error(msg)
            raise ValueError(msg)
        self.var_to_profile = var_to_profile

    def set_profile_grid(self, profile_grid: np.array):
        """
        Parameters
        ----------
        scan_grid
            values of the variable we are going to scan over to compute test-statistics at
        """

        self.profile_grid = profile_grid

    def wilks_ts_crit(self, CL: float) -> float:
        """
        Parameters
        ----------
        CL
            The confidence level at which to compute the critical value, i.e. 0.9 for 90%

        Returns
        -------
        t_crit
            The critical value of the test statistic

        Notes
        -----
        Using Wilks' approximation, we assume the test statistic has a chi-square PDF with 1 degree of freedom.
        We compute the critical value of the test statistic for the given confidence level.
        This is independent of the parameter we are scanning over.
        """
        alpha = 1 - CL  # convert to a significance level, an alpha, one-sided

        return 2 * erfcinv(alpha) ** 2

    def scan_ts(
        self, var_values: np.array, profile_dict: dict = {}  # noqa:B006
    ) -> np.array:
        """
        Parameters
        ----------
        var_values
            The values of the variable to scan over
        profile_dict
            Other values of variables the user wants fixed during a scan

        Returns
        -------
        ts
            Value of the specified test statistic at the scanned values
        """
        # Create the arguments to multiprocess over
        args = [
            [{f"{self.var_to_profile}": float(xx), **profile_dict}] for xx in var_values
        ]

        with mp.Pool(self.numcores) as pool:
            ts = pool.starmap(self.ts, args)
        return ts

    def toy_ts_mp(
        self,
        parameters: dict,  # parameters and values needed to generate the toys
        profile_parameters: dict,  # which parameters to fix and their value (rest are profiled)
        num: int = 1,
    ):
        """
        Makes a number of toys and returns their test statistics. Multiprocessed
        """
        x = np.arange(0, num)
        toys_per_core = np.full(self.numcores, num // self.numcores)
        toys_per_core = np.insert(
            toys_per_core, len(toys_per_core), num % self.numcores
        )

        # remove any cores with 0 toys
        index = np.argwhere(toys_per_core == 0)
        toys_per_core = np.delete(toys_per_core, index)

        # In order to ensure toys aren't correlated between experiments, use the experiment name to set the seed

        experiment_seed = 0

        for c in self.name:
            experiment_seed += ord(c)

        if experiment_seed > 2**31:
            raise ValueError(
                "Experiment seed cannot be too large, try naming the experiment a smaller string."
            )

        # Pick the random seeds that we will pass to toys
        seeds = np.arange(
            experiment_seed + self.jobid * self.numtoy,
            experiment_seed + (self.jobid + 1) * self.numtoy,
        )
        seeds *= 5000  # need to multiply this by a large number because if seed numbers differ by fewer than num of datasets, then adjacent toys will have the same energies pulled but in different datasets
        # See line 115 in toys.py, thisseed = self.seed + i
        # If you have more than 5000 datasets, I am sorry
        if len(self.datasets.items()) > 5000:
            raise ValueError(
                "You need to change the spacing between seeds for completely uncorrelated toys."
            )

        if (seeds > 2**32).any():
            raise ValueError(
                "Experiment seed cannot be too large, try multiplying the seeds by a smaller number."
            )
        seeds_per_toy = []

        j = 0
        for i, num in enumerate(toys_per_core):
            seeds_per_toy.append(seeds[j : j + num])
            j = j + num

        args = [
            [parameters, profile_parameters, num_toy, seeds_per_toy[i]]
            for i, num_toy in enumerate(toys_per_core)
        ]  # give each core multiple MCs

        with mp.Pool(self.numcores) as pool:
            return_args = pool.starmap(self.toy_ts, args)

        ts = [arr[0] for arr in return_args]
        data_to_return = [arr[1] for arr in return_args]
        nuisance_to_return = [arr[2] for arr in return_args]
        num_drawn_to_return = [arr[3] for arr in return_args]
        ts_denom = [arr[4] for arr in return_args]
        ts_num = [arr[5] for arr in return_args]

        # data_to_return is a jagged list, each element is a 2d-array filled it nans
        # First, find the maximum length of array we will need to pad to
        maxlen = np.amax([len(arr[0]) for arr in data_to_return])
        data_flattened = [e for arr in data_to_return for e in arr]

        # Need to flatten the data_to_return in order to save it in h5py
        data_to_return_flat = np.ones((len(data_flattened), maxlen)) * np.nan
        for i, arr in enumerate(data_flattened):
            data_to_return_flat[i, : len(arr)] = arr

        maxlen = np.amax([len(arr[0]) for arr in num_drawn_to_return])
        num_drawn_flattened = [e for arr in num_drawn_to_return for e in arr]

        # Need to flatten the data_to_return in order to save it in h5py
        num_drawn_to_return_flat = np.ones((len(num_drawn_flattened), maxlen)) * np.nan
        for i, arr in enumerate(num_drawn_flattened):
            num_drawn_to_return_flat[i, : len(arr)] = arr

        return (
            np.hstack(ts),
            data_to_return_flat,
            np.vstack(nuisance_to_return),
            num_drawn_to_return_flat,
            seeds,
            np.hstack(ts_denom),
            np.hstack(ts_num),
        )

    def run_and_save_toys(
        self,
        scan_point,
        profile_dict: dict = {},  # noqa:B006
        scan_point_override=None,
        overwrite_files: bool = None,
        compute_conditional: bool = False,
        save_only_ts: bool = True,
        compute_coverage: bool = False,
        toy_pars_override: dict = None,  # noqa:B006
    ):
        """
        Runs toys at specified scan point.
        This can be used to scan on a hypercube if `profile_dict` is passed and `compute_conditional` is false,
        as `profile_dict` is used to fix parameters when generating the toys as well as to fix parameters during the profile.

        Parameters
        ----------
        profile_dict
            An optional dictionary of values we want to fix during all of the profiles

        overwrite_files
            whether to overwrite result files if found, uses global option of SetLimit as default

        compute_conditional
            If true, `profile_dict` is passed, then toys are generated at those values, but allowed to float during the profile fit

        save_only_ts
            Save only the test statistics

        compute_coverage
            If true, use `self.profile_grid` to scan all points on the profile grid, not just scan point. Useful for determining coverage
            by computing limits toy-by-toy and comparing limit to `scan_point`

        toy_pars_override
            If a dict is passed, generate toys at these values instead of the ones profiled out at `scan_point`
        """

        if overwrite_files is None:
            overwrite_files = self.overwrite_files

        filename = ""
        if not profile_dict:
            filename = self.out_path + f"/{scan_point}_{self.jobid}.h5"
        else:
            filename = (
                self.out_path
                + f"/{scan_point}_{list(profile_dict.values())}_{self.jobid}.h5"
            )

        if os.path.exists(filename) and not overwrite_files:
            msg = f"file {filename} exists - use option `overwrite_files` to overwrite"
            raise RuntimeError(msg)

        # First we need to pass the parameter values to generate the toys at: either profile out the variable we are scanning, or user supplied
        if toy_pars_override:
            toypars = toy_pars_override
        else:
            toypars = self.profile(
                {f"{self.var_to_profile}": scan_point, **profile_dict}
            )["values"]

        if scan_point_override is not None:
            toypars[f"{self.var_to_profile}"] = scan_point_override
        else:
            toypars[
                f"{self.var_to_profile}"
            ] = scan_point  # override here if we want to compare the power of the toy ts to another scan_point

        # Now we can run the toys
        if compute_conditional and profile_dict:
            # don't fix the profile_dict points during the fits
            # compute the test statistic at just the given point by default

            (
                toyts,
                data,
                nuisance,
                num_drawn,
                seeds_to_save,
                toyts_denom,
                toyts_num,
            ) = self.toy_ts_mp(
                toypars,
                {f"{self.var_to_profile}": scan_point},
                num=self.numtoy,
            )
        else:
            # Scan a toy over all points if asked
            if compute_coverage and self.profile_grid is not None:
                # check if the profile grid has the scan point in it, otherwise fail gracefully
                if scan_point not in self.profile_grid:
                    raise ValueError(
                        f"{scan_point} not in user specified grid {self.profile_grid}!"
                    )
                (
                    toyts,
                    data,
                    nuisance,
                    num_drawn,
                    seeds_to_save,
                    toyts_denom,
                    toyts_num,
                ) = self.toy_ts_mp(
                    toypars,
                    [
                        {f"{self.var_to_profile}": one_scan_point}
                        for one_scan_point in self.profile_grid
                    ],
                    num=self.numtoy,
                )
            # Default to scanning over just the one point specified
            else:
                (
                    toyts,
                    data,
                    nuisance,
                    num_drawn,
                    seeds_to_save,
                    toyts_denom,
                    toyts_num,
                ) = self.toy_ts_mp(
                    toypars,
                    {f"{self.var_to_profile}": scan_point, **profile_dict},
                    num=self.numtoy,
                )

        # Now, save the toys to a file

        if overwrite_files and os.path.exists(filename):
            msg = f"overwriting existing file {filename}"
            logging.warning(msg)
            os.remove(filename)

        f = h5py.File(filename, "a")

        if profile_dict:
            dset = f.create_dataset(
                "profile_parameters_names", data=list(profile_dict.keys())
            )
            dset = f.create_dataset(
                "profile_parameters_values", data=list(profile_dict.values())
            )

        dset = f.create_dataset("ts", data=toyts)
        dset = f.create_dataset("s", data=scan_point)

        if not save_only_ts:
            dset = f.create_dataset("ts_denom", data=toyts_denom)
            dset = f.create_dataset("ts_num", data=toyts_num)
            dset = f.create_dataset("s", data=scan_point)
            dset = f.create_dataset("Es", data=data)
            # dset = f.create_dataset("nuisance", data=nuisance)
            dset = f.create_dataset("num_sig_num_bkg_drawn", data=num_drawn)
            dset = f.create_dataset("seed", data=seeds_to_save)

        f.close()

        return None

    def run_and_save_brazil(
        self,
        scan_points: list,
        profile_dict: dict = {},  # noqa:B006
        overwrite_files: bool = None,
        compute_conditional: bool = False,
        save_only_ts: bool = True,
        toy_pars_override: dict = None,
    ) -> None:
        """
        Runs toys at 0 signal rate and computes the test statistic for different signal hypotheses.
        If compute_conditional is True and profile_dict is passed, then toys are generated at `profile_dict`,
        but are allowed to float in the fits.

        If compute_conditional is passed as False and a `profile_dict` is passed in order to scan a hypercube, it will work.
        However, this will be slow, and the recommended job submission differs. Use `run_and_save_brazil_with_profile_parameters` instead.

        save_only_ts
            Save only the test statistics

        toy_pars_override
            If a dict is passed, generate toys at these values instead of the ones profiled out at `scan_point`
        """

        if profile_dict and not compute_conditional:
            log.warning(
                "Job submission for hypercube scan using `profile_dict` and `compute_conditional` as False is not recommended.\
                 Use `run_and_save_brazil_with_profile_parameters` instead "
            )

        if overwrite_files is None:
            overwrite_files = self.overwrite_files

        filename = ""
        if not profile_dict:
            filename = self.out_path + f"/0_{self.jobid}.h5"
        else:
            filename = (
                self.out_path + f"/0_{list(profile_dict.values())}_{self.jobid}.h5"
            )

        if os.path.exists(filename) and not overwrite_files:
            msg = f"file {filename} exists - use option `overwrite_files` to overwrite"
            raise RuntimeError(msg)

        # First we need to pass the parameter values to generate the toys at:  either profile out the variable we are scanning at 0 signal rate, or user supplied
        if toy_pars_override:
            toypars = toy_pars_override
        else:
            toypars = self.profile({f"{self.var_to_profile}": 0.0, **profile_dict})[
                "values"
            ]

        # Add 0 to the scan points if it is not there
        if 0.0 not in scan_points:
            scan_points = np.insert(scan_points, 0, 0.0)

        # Now we can run the toys
        if compute_conditional and profile_dict:
            (
                toyts,
                data,
                nuisance,
                num_drawn,
                seeds_to_save,
                toyts_denom,
                toyts_num,
            ) = self.toy_ts_mp(
                toypars,
                [{f"{self.var_to_profile}": scan_point} for scan_point in scan_points],
                num=self.numtoy,
            )
        else:
            (
                toyts,
                data,
                nuisance,
                num_drawn,
                seeds_to_save,
                toyts_denom,
                toyts_num,
            ) = self.toy_ts_mp(
                toypars,
                [
                    {f"{self.var_to_profile}": scan_point, **profile_dict}
                    for scan_point in scan_points
                ],
                num=self.numtoy,
            )

        # Now, save the toys to a file
        if overwrite_files and os.path.exists(filename):
            msg = f"overwriting existing file {filename}"
            logging.warning(msg)
            os.remove(filename)

        f = h5py.File(filename, "a")
        dset = f.create_dataset("ts", data=toyts)
        dset = f.create_dataset("s", data=scan_points)

        if not save_only_ts:
            dset = f.create_dataset("ts_num", data=toyts_num)
            dset = f.create_dataset("ts_denom", data=toyts_denom)
            dset = f.create_dataset("Es", data=data)
            dset = f.create_dataset("nuisance", data=nuisance)
            dset = f.create_dataset("num_sig_num_bkg_drawn", data=num_drawn)
            dset = f.create_dataset("seed", data=seeds_to_save)

        f.close()

        return None

    def run_and_save_brazil_with_profile_parameters(
        self,
        scan_point: float,
        profile_dict: dict = {},  # noqa:B006
        overwrite_files: bool = None,
        compute_conditional: bool = False,
        save_only_ts: bool = True,
    ) -> None:
        """
        Runs toys at 0 signal rate and computes the test statistic for different signal hypotheses
        If we are running with profile_dict parameters in order to scan a hypercube,
        the optimal job submission differs from the above and is more similar to run_and_save_toys

        TODO: refactor this with `run_and_save_brazil`
        """

        if overwrite_files is None:
            overwrite_files = self.overwrite_files

        filename = ""
        if not profile_dict:
            filename = self.out_path + f"/0_{self.jobid}.h5"
        else:
            filename = (
                self.out_path + f"/0_{list(profile_dict.values())}_{self.jobid}.h5"
            )

        if os.path.exists(filename) and not overwrite_files:
            msg = f"file {filename} exists - use option `overwrite_files` to overwrite"
            raise RuntimeError(msg)

        # First we need to profile out the variable we are scanning at 0 signal rate
        toypars = self.profile({f"{self.var_to_profile}": 0.0, **profile_dict})[
            "values"
        ]

        # Now we can run the toys
        if compute_conditional and profile_dict:
            # don't fix the profile_dict points during the fits
            (
                toyts,
                data,
                nuisance,
                num_drawn,
                seeds_to_save,
                toyts_denom,
                toyts_num,
            ) = self.toy_ts_mp(
                toypars,
                {f"{self.var_to_profile}": scan_point},
                num=self.numtoy,
            )
        else:
            (
                toyts,
                data,
                nuisance,
                num_drawn,
                seeds_to_save,
                toyts_denom,
                toyts_num,
            ) = self.toy_ts_mp(
                toypars,
                {f"{self.var_to_profile}": scan_point, **profile_dict},
                num=self.numtoy,
            )

        # Now, save the toys to a file
        if overwrite_files and os.path.exists(filename):
            msg = f"overwriting existing file {filename}"
            logging.warning(msg)
            os.remove(filename)

        f = h5py.File(filename, "a")
        dset = f.create_dataset("ts", data=toyts)
        dset = f.create_dataset("seed", data=seeds_to_save)
        dset = f.create_dataset(
            "profile_parameters_names", data=list(profile_dict.keys())
        )
        dset = f.create_dataset(
            "profile_parameters_values", data=list(profile_dict.values())
        )
        dset = f.create_dataset("s", data=scan_point)

        if not save_only_ts:
            dset = f.create_dataset("ts_num", data=toyts_num)
            dset = f.create_dataset("ts_denom", data=toyts_denom)
            dset = f.create_dataset("Es", data=data)
            dset = f.create_dataset("nuisance", data=nuisance)
            dset = f.create_dataset("num_sig_num_bkg_drawn", data=num_drawn)

        f.close()

        return None

    def run_and_save_joint_profile_toys(
        self,
        scan_points: np.array,
        profile_dict: dict = {},  # noqa:B006
        scan_point_override=None,
        overwrite_files: bool = None,
        compute_conditional: bool = False,
        save_only_ts: bool = True,
    ):
        """
        Runs toys at list of scan points.
        This can be used to scan on a hypercube if `profile_dict` is passed and `compute_conditional` is false,
        as `profile_dict` is used to fix parameters when generating the toys as well as to fix parameters during the profile.
        This differs from `run_and_save_toys` because `scan_points` is now an array.

        TODO: combine this and `run_and_save_toys`

        Parameters
        ----------
        profile_dict
            An optional dictionary of values we want to fix during all of the profiles

        overwrite_files
            whether to overwrite result files if found, uses global option of SetLimit as default

        compute_conditional
            If true, `profile_dict` is passed, then toys are generated at those values, but allowed to float during the profile fit

        save_only_ts
            Save only the test statistics
        """

        if overwrite_files is None:
            overwrite_files = self.overwrite_files

        for scan_point in scan_points:
            filename = ""
            if not profile_dict:
                filename = self.out_path + f"/{scan_point}_{self.jobid}.h5"
            else:
                filename = (
                    self.out_path
                    + f"/{scan_point}_{list(profile_dict.values())}_{self.jobid}.h5"
                )

            if os.path.exists(filename) and not overwrite_files:
                msg = f"file {filename} exists - use option `overwrite_files` to overwrite"
                raise RuntimeError(msg)

            # First we need to profile out the variable we are scanning
            toypars = self.profile(
                {f"{self.var_to_profile}": scan_point, **profile_dict}
            )["values"]

            if scan_point_override is not None:
                toypars[f"{self.var_to_profile}"] = scan_point_override
            else:
                toypars[
                    f"{self.var_to_profile}"
                ] = scan_point  # override here if we want to compare the power of the toy ts to another scan_point

            # Now we can run the toys
            if compute_conditional and profile_dict:
                # don't fix the profile_dict points during the fits
                (
                    toyts,
                    data,
                    nuisance,
                    num_drawn,
                    seeds_to_save,
                    toyts_denom,
                    toyts_num,
                ) = self.toy_ts_mp(
                    toypars,
                    {f"{self.var_to_profile}": scan_point},
                    num=self.numtoy,
                )
            else:
                (
                    toyts,
                    data,
                    nuisance,
                    num_drawn,
                    seeds_to_save,
                    toyts_denom,
                    toyts_num,
                ) = self.toy_ts_mp(
                    toypars,
                    {f"{self.var_to_profile}": scan_point, **profile_dict},
                    num=self.numtoy,
                )

            # Now, save the toys to a file

            if overwrite_files and os.path.exists(filename):
                msg = f"overwriting existing file {filename}"
                logging.warning(msg)
                os.remove(filename)

            f = h5py.File(filename, "a")

            if profile_dict:
                dset = f.create_dataset(
                    "profile_parameters_names", data=list(profile_dict.keys())
                )
                dset = f.create_dataset(
                    "profile_parameters_values", data=list(profile_dict.values())
                )

            dset = f.create_dataset("ts", data=toyts)
            dset = f.create_dataset("s", data=scan_point)

            if not save_only_ts:
                dset = f.create_dataset("ts_denom", data=toyts_denom)
                dset = f.create_dataset("ts_num", data=toyts_num)
                dset = f.create_dataset("Es", data=data)
                # dset = f.create_dataset("nuisance", data=nuisance)
                dset = f.create_dataset("num_sig_num_bkg_drawn", data=num_drawn)
                dset = f.create_dataset("seed", data=seeds_to_save)

            f.close()

        return None



import matplotlib.pyplot as plt
import numba as nb
import numpy as np

import freqfit.models.constants as constants
from freqfit.utils import inspectparameters

nb_kwd = {
    "nopython": True,
    "parallel": False,
    "nogil": True,
    "cache": True,
    "fastmath": True,
    "inline": "always",
}

QBB = constants.QBB
N_A = constants.NA
M_A = constants.MDET

# default analysis window and width
WINDOW = np.array(constants.WINDOW)

WINDOWSIZE = 0.0
for i in range(len(WINDOW)):
    WINDOWSIZE += WINDOW[i][1] - WINDOW[i][0]

SEED = 42  # set the default random seed


@nb.jit(**nb_kwd)
def nb_pdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    exp
        The exposure, in kg*yr
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function computes the following:
    mu_S = eff * exp * S
    mu_B = exp * BI * windowsize
    pdf(E) = 1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * eff * exp
    mu_B = exp * BI * WINDOWSIZE

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB - delta) ** 2) / (2 * sigma**2)) + B_amp
        )

    return y


@nb.jit(**nb_kwd)
def nb_density(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    exp
        The exposure, in kg*yr
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function computes the following, faster than without a numba wrapper:
    mu_S = eff * exp * S
    mu_B = exp * BI * windowsize
    CDF(E) = mu_S + mu_B
    pdf(E) =[mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize]
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * eff * exp
    mu_B = exp * BI * WINDOWSIZE

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        y[i] = S_amp * np.exp(-((Es[i] - QBB - delta) ** 2) / (2 * sigma**2)) + B_amp

    return mu_S + mu_B, y


@nb.jit(**nb_kwd)
def nb_logpdf(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    exp
        The exposure, in kg*yr
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function computes the following:
    mu_S = eff * exp * S
    mu_B = exp * BI * windowsize
    logpdf(E) = log(1/(mu_S+mu_B) * [mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize])
    """

    # Precompute the signal and background counts
    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * eff * exp
    mu_B = exp * BI * WINDOWSIZE

    if sigma == 0:  # need this check for fitting
        return np.full_like(
            Es, np.log(exp * BI / (mu_S + mu_B))
        )  # TODO: make sure this simplification makes sense in the limit

    # Precompute the prefactors so that way we save multiplications in the for loop
    B_amp = exp * BI
    S_amp = mu_S / (np.sqrt(2 * np.pi) * sigma)

    # Initialize and execute the for loop
    y = np.empty_like(Es, dtype=np.float64)
    for i in nb.prange(Es.shape[0]):
        pdf = (1 / (mu_S + mu_B)) * (
            S_amp * np.exp(-((Es[i] - QBB - delta) ** 2) / (2 * sigma**2)) + B_amp
        )

        if pdf <= 0:
            y[i] = -np.inf
        else:
            y[i] = np.log(pdf)

    return y


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_rvs(
    n_sig: int,
    n_bkg: int,
    delta: float,
    sigma: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    n_sig
        Number of signal events to pull from
    n_bkg
        Number of background events to pull from
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    seed
        specify a seed, otherwise uses default seed
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.
    """

    np.random.seed(seed)

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB + delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es


@nb.jit(nopython=True, fastmath=True, cache=True, error_model="numpy")
def nb_extendedrvs(
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    exp: float,
    seed: int = SEED,
) -> np.array:
    """
    Parameters
    ----------
    S
        expected rate of signal events in events/(kg*yr)
    BI
        rate of background events in events/(kev*kg*yr)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    seed
        specify a seed, otherwise uses default seed
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function pulls from a Gaussian for signal events and from a uniform distribution for background events
    in the provided windows, which may be discontinuous.

    Returns
    -------
    rvs
        The random variables drawn
    (n_sig, n_bkg)
        Number of signal counts drawn and number of background counts
    """

    np.random.seed(seed)

    n_sig = np.random.poisson(S * eff * exp)
    n_bkg = np.random.poisson(BI * exp * WINDOWSIZE)

    # Get energy of signal events from a Gaussian distribution
    # preallocate for background draws
    Es = np.append(np.random.normal(QBB + delta, sigma, size=n_sig), np.zeros(n_bkg))

    # Get background events from a uniform distribution
    bkg = np.random.uniform(0, 1, n_bkg)

    breaks = np.zeros(shape=(len(WINDOW), 2))
    for i in range(len(WINDOW)):
        thiswidth = WINDOW[i][1] - WINDOW[i][0]

        if i > 0:
            breaks[i][0] = breaks[i - 1][1]

        if i < len(WINDOW):
            breaks[i][1] = breaks[i][0] + thiswidth / WINDOWSIZE

        for j in range(len(bkg)):
            if breaks[i][0] <= bkg[j] <= breaks[i][1]:
                Es[n_sig + j] = (bkg[j] - breaks[i][0]) * thiswidth / (
                    breaks[i][1] - breaks[i][0]
                ) + WINDOW[i][0]

    return Es, (n_sig, n_bkg)


@nb.jit(**nb_kwd)
def nb_density_gradient(
    Es: np.array,
    S: float,
    BI: float,
    delta: float,
    sigma: float,
    eff: float,
    exp: float,
) -> np.array:
    """
    Parameters
    ----------
    Es
        Energies at which this function is evaluated, in keV
    S
        The signal rate, in units of counts/(kg*yr)
    BI
        The background index rate, in counts/(kg*yr*keV)
    delta
        Systematic energy offset from QBB, in keV
    sigma
        The energy resolution at QBB, in keV
    eff
        The global signal efficiency, unitless
    exp
        The exposure, in kg*yr
    window
        uniform background regions to pull from, must be a 2D array of form e.g. `np.array([[0,1],[2,3]])`
        where edges of window are monotonically increasing (this is not checked), in keV.
        Default is typical analysis window.

    Notes
    -----
    This function computes the gradient of the density function and returns a tuple where the first element is the gradient of the CDF, and the second element is the gradient of the PDF.
    The first element has shape (K,) where K is the number of parameters, and the second element has shape (K,N) where N is the length of Es.
    mu_S = eff * exp * S
    mu_B = exp * BI * windowsize
    pdf(E) = [mu_S * norm(E_j, QBB + delta, sigma) + mu_B/windowsize]
    cdf(E) = mu_S + mu_B
    """

    # mu_S = np.log(2) * (N_A * S) * eff * exp / M_A
    mu_S = S * eff * exp

    grad_CDF = np.array(
        [eff * exp, exp * WINDOWSIZE, 0, 0, S * exp, S * eff + BI * WINDOWSIZE]
    )

    grad_PDF = np.zeros(shape=(6, len(Es)))
    for i in nb.prange(Es.shape[0]):
        # For readability, don't precompute anything and see how performance is impacted
        grad_PDF[0][i] = (
            eff
            * exp
            * (1 / (np.sqrt(2 * np.pi) * sigma))
            * np.exp(-1 * (Es[i] - QBB - delta) ** 2 / (2 * sigma**2))
        )
        grad_PDF[1][i] = exp
        grad_PDF[2][i] = (
            ((Es[i] - QBB - delta) / sigma**2)
            * mu_S
            * (1 / (np.sqrt(2 * np.pi) * sigma))
            * np.exp(-1 * (Es[i] - QBB - delta) ** 2 / (2 * sigma**2))
        )
        grad_PDF[3][i] = (
            (
                ((Es[i] - QBB - delta) ** 2 - sigma**2)
                / (np.sqrt(2 * np.pi) * sigma**4)
            )
            * mu_S
            * np.exp(-1 * (Es[i] - QBB - delta) ** 2 / (2 * sigma**2))
        )
        grad_PDF[4][i] = (
            S
            * exp
            * (1 / (np.sqrt(2 * np.pi) * sigma))
            * np.exp(-1 * (Es[i] - QBB - delta) ** 2 / (2 * sigma**2))
        )
        grad_PDF[5][i] = (
            S
            * eff
            * (1 / (np.sqrt(2 * np.pi) * sigma))
            * np.exp(-1 * (Es[i] - QBB - delta) ** 2 / (2 * sigma**2))
            + BI
        )

    return grad_CDF, grad_PDF


class gaussian_on_uniform_gen:
    def __init__(self):
        self.parameters = inspectparameters(self.density)
        pass

    def pdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> np.array:
        return nb_pdf(Es, S, BI, delta, sigma, eff, exp)

    def logpdf(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> np.array:
        return nb_logpdf(Es, S, BI, delta, sigma, eff, exp)

    # for iminuit ExtendedUnbinnedNLL
    def density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> np.array:
        return nb_density(Es, S, BI, delta, sigma, eff, exp)

    def density_gradient(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> tuple:
        return nb_density_gradient(Es, S, BI, delta, sigma, eff, exp)

    # for iminuit ExtendedUnbinnedNLL
    def log_density(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> np.array:
        mu_S = S * eff * exp
        mu_B = exp * BI * WINDOWSIZE

        # Do a quick check and return -inf if log args are negative
        if (mu_S + mu_B <= 0) or np.isnan(np.array([mu_S, mu_B])).any():
            return mu_S + mu_B, np.full(Es.shape[0], -np.inf)
        else:
            return (
                mu_S + mu_B,
                np.log(mu_S + mu_B) + nb_logpdf(Es, S, BI, delta, sigma, eff, exp),
            )

    # should we have an rvs method for drawing a random number of events?
    # `extendedrvs`
    # needs to use same parameters as the rest of the functions...
    def rvs(
        self,
        n_sig: int,
        n_bkg: int,
        delta: float,
        sigma: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_rvs(n_sig, n_bkg, delta, sigma, seed=seed)

    def extendedrvs(
        self,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
        seed: int = SEED,
    ) -> np.array:
        return nb_extendedrvs(S, BI, delta, sigma, eff, exp, seed=seed)

    def plot(
        self,
        Es: np.array,
        S: float,
        BI: float,
        delta: float,
        sigma: float,
        eff: float,
        exp: float,
    ) -> None:
        y = nb_pdf(Es, S, BI, delta, sigma, eff, exp)

        plt.step(Es, y)
        plt.show()


gaussian_on_uniform = gaussian_on_uniform_gen()
